//
// Created by philipp on 30.11.17.
//

#include <iostream>
#include "../include/PlatoonProtocolLib/LeaderVehicle.h"
#include "../include/PlatoonProtocolLib/Utils.h"

using namespace platoonProtocol;
using namespace std::chrono_literals;

constexpr networking::time::Duration LeaderVehicle::HEARTBEAT_TIMEOUT;
constexpr networking::time::Duration LeaderVehicle::BROADCAST_INTERVAL;

LeaderVehicle::Ptr LeaderVehicle::create(networking::Networking & net, VehicleId ownVehicleId)
{
    return std::make_shared<LeaderVehicle>(PrivateTag{}, net, ownVehicleId);
}

LeaderVehicle::LeaderVehicle(PrivateTag, networking::Networking & net, VehicleId ownVehicleId)
    : Vehicle(net, ownVehicleId, Role::LEADER)
{
    broadcastTimer = networking::time::Timer::create(net);
    newHeartbeatReceiver();
}

void LeaderVehicle::stop()
{
    Vehicle::stop();

    removeAllFollowers();

    stopBroadcasting();

    stopReceiveHeartbeats();
}

void LeaderVehicle::setPlatoonConfig(const PlatoonConfig & platoonConfig)
{
    this->platoonConfig = platoonConfig;
}

void LeaderVehicle::setPlatoonSpeed(PlatoonSpeed platoonSpeed)
{
    platoonConfig = PlatoonConfig{platoonSpeed, platoonConfig.get().innerPlatoonDistance};
}

void LeaderVehicle::setInnerPlatoonDistance(InnerPlatoonDistance innerPlatoonDistance)
{
    platoonConfig = PlatoonConfig{platoonConfig.get().platoonSpeed, innerPlatoonDistance};
}

std::size_t LeaderVehicle::getNumFollowers()
{
    std::lock_guard<std::mutex> lock(followersMutex);
    return followers.size();
}

void LeaderVehicle::doCreatePlatoon()
{
    if (state != State::IDLE)
        return;

    setState(State::CREATING_PLATOON);

    // Generate a platoon id.
    platoonId = rand();

    Vehicle::doCreatePlatoon();
}

void LeaderVehicle::doRunPlatoon()
{
    if (state != State::CREATING_PLATOON)
        return;

    setState(State::RUNNING_PLATOON);

    startBroadcasting();

    startReceiveHeartbeats();

    onRunningPlatoonCallback();
}

void LeaderVehicle::doLeavePlatoon()
{
    stop();

    if (state == State::RUNNING_PLATOON)
        onLeavingPlatoonCallback();

    setState(State::IDLE);
}

void LeaderVehicle::makePlatoonCreateRequest(const VehicleEndpoint & vehicleEndpoint)
{
    // We'll send a request indicating that we want to be the leader.
    auto self = shared_from_this();
    newPlatoonCreateServiceClient();
    platoonCreateServiceClient->asyncCall(
        PlatoonMessage::leaderRequest(getOwnEndpoint().getVehicleId(), platoonId),
        vehicleEndpoint.getHost(),
        vehicleEndpoint.getTcpPort(),
        3s,
        [self, vehicleEndpoint](const auto & error, const auto & response)
        {
            if (self->state != State::CREATING_PLATOON &&
                self->state != State::RUNNING_PLATOON)
                return;

            if (error)
                return;

            if (response.vehicleId != vehicleEndpoint.getVehicleId() ||
                response.messageType != messageTypes::ACCEPT_RESPONSE ||
                response.platoonId != self->platoonId)
                return;

            self->addFollower(vehicleEndpoint);

            self->doRunPlatoon();
        });
}

void LeaderVehicle::makePlatoonCreateResponse(const networking::Resolver::Endpoint & clientEndpoint,
                                              const PlatoonMessage & request,
                                              PlatoonMessage & response)
{
    if (state != State::CREATING_PLATOON &&
        state != State::RUNNING_PLATOON)
    {
        response = PlatoonMessage::rejectResponse(getOwnEndpoint().getVehicleId());
        return;
    }

    auto vehicleId = request.vehicleId;
    // Only consider this message if it is a follower request,
    // meaning that a follower wants us to be its leader.
    if (request.messageType != messageTypes::FV_REQUEST || vehicleId == 0)
    {
        response = PlatoonMessage::rejectResponse(getOwnEndpoint().getVehicleId());
        return;
    }

    addFollower(VehicleEndpoint{request.vehicleId, clientEndpoint.ip});

    doRunPlatoon();

    response = PlatoonMessage::acceptResponse(getOwnEndpoint().getVehicleId(), platoonId);
}

bool LeaderVehicle::findFollower(VehicleId id)
{
    return followers.find(id) != followers.end();
}

void LeaderVehicle::addFollower(const VehicleEndpoint & endpoint)
{
    if (findFollower(endpoint.getVehicleId()))
        return;

    Follower follower;
    follower.endpoint = endpoint;
    follower.broadcastSender = networking::message::DatagramSender<PlatoonMessage>::create(net);
    follower.heartbeatTimer = networking::time::Timer::create(net);

    {
        std::lock_guard<std::mutex> lock(followersMutex);
        followers[endpoint.getVehicleId()] = follower;
    }

    startHeartbeatTimer(followers[endpoint.getVehicleId()]);
}

void LeaderVehicle::startHeartbeatTimer(Follower & follower)
{
    auto followerId = follower.endpoint.getVehicleId();
    auto self = shared_from_this();
    follower.heartbeatTimer = networking::time::Timer::create(net);
    follower.heartbeatTimer->startTimeout(
        HEARTBEAT_TIMEOUT,
        [self, followerId]
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            if (!self->findFollower(followerId))
                return;

            auto & follower = self->followers[followerId];
            self->removeFollower(follower);
        });
}

void LeaderVehicle::stopHeartbeatTimer(LeaderVehicle::Follower & follower)
{
    follower.heartbeatTimer->stop();
}

void LeaderVehicle::startReceiveHeartbeats()
{
    newHeartbeatReceiver();
    receiveHeartbeat();
}

void LeaderVehicle::receiveHeartbeat()
{
    auto self = shared_from_this();
    heartbeatReceiver->asyncReceive(
        networking::time::Duration::max(), // No timeout required here.
        [self](const auto & error, const auto & message, const auto & host, auto port)
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            auto followerId = message.vehicleId;

            if (!error &&
                self->findFollower(followerId) &&
                message.platoonId == self->platoonId)
            {
                auto & follower = self->followers[followerId];

                switch (message.messageType)
                {
                    case messageTypes::LEAVE_PLATOON:
                        self->removeFollower(follower);
                        break;

                    case messageTypes::FV_HEARTBEAT:
                        // Reset the timer.
                        self->stopHeartbeatTimer(follower);
                        self->startHeartbeatTimer(follower);
                        break;
                }
            };

            // Receive next message.
            self->receiveHeartbeat();
        });
}

void LeaderVehicle::stopReceiveHeartbeats()
{
    heartbeatReceiver->stop();
}

void LeaderVehicle::startBroadcasting()
{
    auto self = shared_from_this();
    broadcastTimer = networking::time::Timer::create(net);
    broadcastTimer->startPeriodicTimeout(
        BROADCAST_INTERVAL,
        [self]
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            PlatoonConfig configCopy = self->platoonConfig.get();

            for (auto & pair : self->followers)
            {
                auto & follower = pair.second;

                auto broadcastMsg = PlatoonMessage::broadcastMessage(
                    self->getOwnEndpoint().getVehicleId(),
                    self->platoonId,
                    configCopy.platoonSpeed,
                    configCopy.innerPlatoonDistance);

                // Send the broadcast message and forget about it.
                follower.broadcastSender->asyncSend(
                    broadcastMsg, follower.endpoint.getHost(), follower.endpoint.getUdpPort(), BROADCAST_INTERVAL,
                    [self](auto && ...)
                    {});
            }
        });
}

void LeaderVehicle::stopBroadcasting()
{
    broadcastTimer->stop();
}

void LeaderVehicle::removeFollower(Follower & follower)
{
    stopHeartbeatTimer(follower);
    {
        std::lock_guard<std::mutex> lock(followersMutex);
        followers.erase(follower.endpoint.getVehicleId());
    }
    // If there are no more followers, we leave the platoon too.
    if (followers.empty())
        doLeavePlatoon();
}

void LeaderVehicle::removeAllFollowers()
{
    std::lock_guard<std::mutex> lock{followersMutex};
    utils::foreachInMap(
        followers,
        [&](auto & follower)
        {
            this->stopHeartbeatTimer(follower);
            followers.erase(follower.endpoint.getVehicleId());
        });
}

void LeaderVehicle::newHeartbeatReceiver()
{
    heartbeatReceiver = networking::message::DatagramReceiver<PlatoonMessage>::create(
        net, getOwnEndpoint().getUdpPort(), PlatoonMessage::MAX_SIZE);
}

LeaderVehicle::Ptr LeaderVehicle::shared_from_this()
{
    return std::static_pointer_cast<LeaderVehicle>(Vehicle::shared_from_this());
}
